import torch

class AUCMLoss(torch.nn.Module):
    """
    AUCM Loss with squared-hinge function: a novel loss function to directly optimize AUROC
    
    inputs:
        margin: margin term for AUCM loss, e.g., m in [0, 1]
        imratio: imbalance ratio, i.e., the ratio of number of postive samples to number of total samples
    outputs:
        loss value 
    
    Reference: 
        Yuan, Z., Yan, Y., Sonka, M. and Yang, T., 
        Large-scale Robust Deep AUC Maximization: A New Surrogate Loss and Empirical Studies on Medical Image Classification. 
        International Conference on Computer Vision (ICCV 2021)
    Link:
        https://arxiv.org/abs/2012.03173
    """
    def __init__(self, margin=1.0, imratio=None, device=None):
        super(AUCMLoss, self).__init__()
        if not device:
            self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        else:
            self.device = device   
        self.margin = margin
        self.p = imratio
        # https://discuss.pytorch.org/t/valueerror-cant-optimize-a-non-leaf-tensor/21751
        self.a = torch.zeros(1, dtype=torch.float32, device=self.device, requires_grad=True).to(self.device) 
        self.b = torch.zeros(1, dtype=torch.float32, device=self.device,  requires_grad=True).to(self.device) 
        self.alpha = torch.zeros(1, dtype=torch.float32, device=self.device, requires_grad=True).to(self.device) 
        
    def forward(self, y_pred, y_true):
        if self.p is None:
           self.p = (y_true==1).float().sum()/y_true.shape[0]   
     
        y_pred = y_pred.reshape(-1, 1) # be carefull about these shapes
        y_true = y_true.reshape(-1, 1) 
        loss = (1-self.p)*torch.mean((y_pred - self.a)**2*(1==y_true).float()) + \
                    self.p*torch.mean((y_pred - self.b)**2*(0==y_true).float())   + \
                    2*self.alpha*(self.p*(1-self.p)*self.margin + \
                    torch.mean((self.p*y_pred*(0==y_true).float() - (1-self.p)*y_pred*(1==y_true).float())) )- \
                    self.p*(1-self.p)*self.alpha**2
        return loss

class AUCNERLoss(AUCMLoss):
    def __init__(self, margin, imratio, device):
        super().__init__(margin, imratio, device)

    def forward(self, y_pred, y_true):
        if self.p is None:
            self.p = (y_true == 1).float().sum() / y_true.shape[0]

        y_pred = y_pred.reshape(-1, 1)  # be carefull about these shapes
        y_true = y_true.reshape(-1, 1)

        loss = (1 - self.p) * ((y_pred - self.a) ** 2 * (1 == y_true).float()) + \
            self.p * ((y_pred - self.b) ** 2 * (0 == y_true).float()) + \
            2 * self.alpha * (self.p * (1 - self.p) * self.margin + ((self.p * y_pred * (0 == y_true).float() - (1 - self.p) * y_pred * (1 == y_true).float()))) - \
            self.p * (1 - self.p) * self.alpha ** 2

        return loss


class OneVsAllAUCNERLoss(torch.nn.Module):
    def __init__(self, args, label):
        super(OneVsAllAUCNERLoss, self).__init__()
        self.args = args
        
        if not args.device:
            self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        else:
            self.device = args.device
        
        self.label = label

        for class_idx in range(args.num_class):
            class_name = self.args.ID_TO_LABEL[class_idx]
            if not class_name.startswith(self.label):
                continue

            self.__setattr__("Loss_" + class_name, AUCNERLoss(
                imratio= args.__getattribute__("imratio_" + class_name), 
                margin=args.margin, device=self.device
            ))


    def forward(self, logits, labels, loss_dict):
        loss = 0
        zeros = torch.zeros_like(labels, device=self.device, dtype=torch.float32)
        mask = ~torch.eq(labels, -1)
        total_mask = sum(mask)
        
        for class_idx in range(self.args.num_class):
            class_name = self.args.ID_TO_LABEL[class_idx]
            if not class_name.startswith(self.label):
                continue
            
            # label_class = torch.eq(labels, class_idx)
            label_class = torch.where(labels==class_idx , 1, 0)
            label_class = torch.where(mask, label_class, -1)
            
            loss_tensors = self.__getattr__("Loss_" + class_name)(torch.sigmoid(logits[class_name]).flatten().unsqueeze(1), label_class.flatten().unsqueeze(1))
            loss_tensors = torch.where(mask.flatten().byte(), loss_tensors, zeros.flatten())

            curr_loss = (torch.sum(loss_tensors) / total_mask)
            
            loss += curr_loss
            
            loss_dict["loss_" + class_name] = curr_loss.item()

        return loss, loss_dict


class OVACEloss(torch.nn.Module):
    def __init__(self, args, label):
        super(OVACEloss, self).__init__()
        self.args = args
        
        if not args.device:
            self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        else:
            self.device = args.device
        
        self.label = label

        for class_idx in range(args.num_class):
            class_name = self.args.ID_TO_LABEL[class_idx]
            if not class_name.startswith(self.label):
                continue

            self.__setattr__("Loss_" + class_name, torch.nn.BCELoss(reduction="none"))

    def forward(self, logits, labels, loss_dict):
        loss = 0
        zeros = torch.zeros_like(labels, device=self.device, dtype=torch.half)
        mask = ~torch.eq(labels, -1)
        total_mask = sum(mask)
        
        for class_idx in range(self.args.num_class):
            class_name = self.args.ID_TO_LABEL[class_idx]
            if not class_name.startswith(self.label):
                continue
            
            # label_class = torch.eq(labels, class_idx)
            label_class = torch.where(labels==class_idx , 1, 0)
            label_class = torch.where(mask, label_class, -1)
            
            loss_tensors = self.__getattr__("Loss_" + class_name)(
                torch.sigmoid(logits[class_name]).flatten().unsqueeze(1).type(torch.half), 
                label_class.flatten().unsqueeze(1).type(torch.half)
            )
            loss_tensors = torch.where(mask.flatten().byte(), loss_tensors, zeros.flatten())

            curr_loss = (torch.sum(loss_tensors) / total_mask)
            
            loss += curr_loss
            
            loss_dict["loss_" + class_name] = curr_loss.item()

        return loss, loss_dict


class MAUCNERLoss(torch.nn.Module):
    def __init__(self, args):
        super(MAUCNERLoss, self).__init__()
        self.args = args
        if not args.device:
            self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        else:
            self.device = args.device
        
        for class_idx in range(args.num_class):
            class_name = self.args.ID_TO_LABEL[class_idx]
            self.__setattr__("Loss_" + class_name, AUCNERLoss(
                imratio= args.__getattribute__("imratio_" + class_name), 
                margin=args.margin, device=self.device
            ))

    def forward(self, logits, labels, labels_avoid=None):
        loss = 0
        zeros = torch.zeros_like(labels, device=self.device, dtype=torch.float32)
        mask = ~torch.eq(labels, -1)
        total_mask = sum(mask)
        loss_dict = {}
        
        for class_idx in range(self.args.num_class):
            class_name = self.args.ID_TO_LABEL[class_idx]

            if labels_avoid is not None:
                if class_name in labels_avoid:
                    continue

            label_class = torch.where(labels==class_idx , 1, 0)
            label_class = torch.where(mask, label_class, -1)
            
            loss_tensors = self.__getattr__("Loss_" + class_name)(
                torch.sigmoid(logits[class_name]).flatten().unsqueeze(1), label_class.flatten().unsqueeze(1)
            )
            loss_tensors = torch.where(mask.flatten().byte(), loss_tensors, zeros.flatten())
            curr_loss = (torch.sum(loss_tensors) / total_mask)
            
            loss += curr_loss
            
            loss_dict["loss_" + class_name] = curr_loss.item()

        return loss, loss_dict