import torch
import torch.nn as nn
from transformers import AutoConfig, AutoModel
from torchcrf import CRF


class NERModel(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args
        config = AutoConfig.from_pretrained(args.model_name_or_path, num_labels=args.num_class)
        self.bertfamily = AutoModel.from_pretrained(args.model_name_or_path,)
        self.dropout = nn.Dropout(args.dropout_prob)
        out_dim = config.hidden_size
        if self.args.bilstmopt:
            self.bilstm = nn.LSTM(config.hidden_size, 128, num_layers=1, bidirectional=True, batch_first=True)
            out_dim = 128*2
        self.classifier = nn.Linear(out_dim, args.num_class)
        self.loss_fnt = nn.CrossEntropyLoss(ignore_index=-1) ## ignore the loss of padding tags with -1
        self.crf = CRF(config.num_labels, batch_first=True)

    def forward(self, batch):
        input_ids = batch['input_ids']
        attention_mask = batch['attention_mask']
        labels = batch['labels']
        h, *_ = self.bertfamily(input_ids, attention_mask, return_dict=False)
        if self.args.bilstmopt:
            self.bilstm.flatten_parameters()
            h, _ = self.bilstm(h)
        c = self.args.num_class
        h = self.dropout(h)
        logits = self.classifier(h) ## 3 dimensions batchsize * seq length * number of class
        logits_flat2d = logits.view(-1, c) ## 2 dimensions (batchsize * seq length) * number of class
        
        if (self.args.crfopt):
            outputs = (logits,)
        else:
            outputs = (logits_flat2d,)

        if labels is not None:
            labels_flat2d = labels.view(-1)
            if (self.args.crfopt):
                mask = ~torch.eq(labels, -1)
                loss = -1 * self.crf(logits[:,1:,], labels[:,1:], mask=mask[:,1:].bool(), reduction="mean")
            else:
                loss = self.loss_fnt(logits_flat2d, labels_flat2d)
            outputs = (loss,) + outputs
        return outputs


class NEROVAModel(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args

        config = AutoConfig.from_pretrained(args.model_name_or_path, num_labels=args.num_class)
        self.bertfamily = AutoModel.from_pretrained(args.model_name_or_path, )
        self.dropout = nn.Dropout(args.dropout_prob)

        out_dim = config.hidden_size
        if self.args.bilstmopt:
            self.bilstm = nn.LSTM(config.hidden_size, 128, num_layers=1, bidirectional=True, batch_first=True)
            out_dim = 128*2

        for i in range(args.num_class):
            class_name = self.args.ID_TO_LABEL[i]
            self.__setattr__("logit_"+class_name, nn.Linear(out_dim, 1))

    def forward(self, batch):
        input_ids = batch['input_ids']
        attention_mask = batch['attention_mask']
        labels = batch['labels']
        labels_en3 = batch['labels_en3']
        labels_en2 = batch['labels_en2']
        labels_be2 = batch['labels_be2']
        h, *_ = self.bertfamily(input_ids, attention_mask, return_dict=False)
        
        if self.args.bilstmopt:
            self.bilstm.flatten_parameters()
            h, _ = self.bilstm(h)
        
        h = self.dropout(h)

        logits = {}

        for i in range(self.args.num_class):
            class_name = self.args.ID_TO_LABEL[i]
            logits[class_name] = self.__getattr__("logit_"+class_name)(h).view(-1,1)
        
        return logits, {"labels":labels, "labels_en3": labels_en3, "labels_en2": labels_en2, "labels_be2" : labels_be2}