import torch
import math
from torch.optim.optimizer import required 


class MPESG(torch.optim.Optimizer):
    def __init__(
        self, 
        model,
        loss,
        args,
        lr,
        gamma=500,
        clip_value=1.0,
        **kwargs
    ):
       
        self.args = args
        for class_idx in range(args.num_class):
            class_name = self.args.ID_TO_LABEL[class_idx]
            self.__setattr__("p_" + class_name, loss.__getattr__("Loss_" + class_name).p)
            self.__setattr__("margin_" + class_name, loss.__getattr__("Loss_" + class_name).margin)
            self.__setattr__("a_" + class_name, loss.__getattr__("Loss_" + class_name).a)
            self.__setattr__("b_" + class_name, loss.__getattr__("Loss_" + class_name).b)
            self.__setattr__("alpha_" + class_name, loss.__getattr__("Loss_" + class_name).alpha)

        self.model = model
        
        if not args.device:
            self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        else:
            self.device = args.device
        
        self.lr = lr
        self.gamma = gamma
        self.clip_value = clip_value
        self.weight_decay = self.args.weight_decay
            
        # TODO! 
        self.model_ref = self.init_model_ref()
        self.model_acc = self.init_model_acc()
        
        self.T = 0
        self.step_counts = 0
    
        def get_parameters(params):
            for p in params:
                yield p
        
        self.params = get_parameters(
            list(model.parameters()) + 
            [self.__getattribute__("a_" + self.args.ID_TO_LABEL[class_idx]) for class_idx in range(self.args.num_class)] +
            [self.__getattribute__("b_" + self.args.ID_TO_LABEL[class_idx]) for class_idx in range(self.args.num_class)]
        )
        
        self.defaults = dict(lr=self.lr, 
                             gamma=gamma, 
                             clip_value=self.clip_value,
                             weight_decay=self.weight_decay,
                             model_ref = self.model_ref,
                             model_acc = self.model_acc
                             )
        
        for class_idx in range(self.args.num_class):
            class_name = self.args.ID_TO_LABEL[class_idx]
            self.defaults["p_" + class_name] = self.__getattribute__("p_" + class_name)
            self.defaults["a_" + class_name] = self.__getattribute__("a_" + class_name)
            self.defaults["b_" + class_name] = self.__getattribute__("b_" + class_name)
            self.defaults["margin_" + class_name] = self.__getattribute__("margin_" + class_name)
            self.defaults["alpha_" + class_name] = self.__getattribute__("alpha_" + class_name)

        super(MPESG, self).__init__(self.params, self.defaults)
     
    def init_model_ref(self):
         self.model_ref = []
         for var in list(self.model.parameters()) + [self.__getattribute__("a_" + self.args.ID_TO_LABEL[class_idx]) for class_idx in range(self.args.num_class)] + [self.__getattribute__("b_" + self.args.ID_TO_LABEL[class_idx]) for class_idx in range(self.args.num_class)]:
            self.model_ref.append(torch.empty(var.shape).normal_(mean=0, std=0.01).to(self.device))
         return self.model_ref
     
    def init_model_acc(self):
        self.model_acc = []
        for var in list(self.model.parameters()) + [self.__getattribute__("a_" + self.args.ID_TO_LABEL[class_idx]) for class_idx in range(self.args.num_class)] + [self.__getattribute__("b_" + self.args.ID_TO_LABEL[class_idx]) for class_idx in range(self.args.num_class)]:
            self.model_acc.append(torch.zeros(var.shape, dtype=torch.float32,  device=self.device, requires_grad=False).to(self.device)) 
        return self.model_acc
    
    @property    
    def optim_steps(self):
        return self.step_counts
    
    @property
    def get_params(self):
        return list(self.model.parameters())
    
    def update_lr(self, lr):
        self.param_groups[0]['lr']=lr

    @torch.no_grad()
    def step(self):
        """Performs a single optimization step.
        """
        for group in self.param_groups:
            weight_decay = group['weight_decay']
            clip_value = group['clip_value']
            self.lr =  group['lr']
            
            gamma = group['gamma']
           
            model_ref = group['model_ref']
            model_acc = group['model_acc']

            # updates
            for i, p in enumerate(group['params']):
                if p.grad is None:
                    continue
                if i >=  (len(group['params']) - self.args.num_class):
                    assert sum(p.size()) == 1
                p.data = p.data - group['lr']*(torch.clamp(p.grad.data , -clip_value, clip_value) + 1/gamma*(p.data - model_ref[i].data) + weight_decay*p.data)
                model_acc[i].data = model_acc[i].data + p.data

            for class_idx in range(self.args.num_class):
                class_name = self.args.ID_TO_LABEL[class_idx]
                group['alpha_' + class_name].data = group['alpha_' + class_name].data + group['lr']*(
                    2*(group['margin_' + class_name] + group['b_' + class_name].data - group['a_' + class_name].data) -
                    2*group['alpha_' + class_name].data
                )
                group['alpha_' + class_name].data = torch.clamp(group['alpha_' + class_name].data, 0, 999)

        self.T += 1  
        self.step_counts += 1

    def zero_grad(self):
        self.model.zero_grad()
        for class_idx in range(self.args.num_class):
            class_name = self.args.ID_TO_LABEL[class_idx]
            self.__getattribute__("a_" + class_name).grad = None
            self.__getattribute__("b_" + class_name).grad = None
            self.__getattribute__("alpha_" + class_name).grad = None
        
    def update_regularizer(self, decay_factor=None):
        if decay_factor != None:
            self.param_groups[0]['lr'] = self.param_groups[0]['lr']/decay_factor
            print ('Reducing learning rate to %.5f @ T=%s!'%(self.param_groups[0]['lr'], self.T))
        print ('Updating regularizer @ T=%s!'%(self.T))
        for i, param in enumerate(self.model_ref):
            self.model_ref[i].data = self.model_acc[i].data/self.T
        for i, param in enumerate(self.model_acc):
            self.model_acc[i].data = torch.zeros(param.shape, dtype=torch.float32, device=self.device,  requires_grad=False).to(self.device)
        self.T = 0


class OneVsAllPESG(torch.optim.Optimizer):
    def __init__(
        self, 
        model,
        loss,
        args,
        label,
        lr,
        gamma=500,
        clip_value=1.0,
        **kwargs
    ):
       
        self.args = args
        self.label = label
        
        for class_idx in range(args.num_class):
            class_name = self.args.ID_TO_LABEL[class_idx]
            if not class_name.startswith(self.label):
                continue
            self.__setattr__("p_" + class_name, loss.__getattr__("Loss_" + class_name).p)
            self.__setattr__("margin_" + class_name, loss.__getattr__("Loss_" + class_name).margin)
            self.__setattr__("a_" + class_name, loss.__getattr__("Loss_" + class_name).a)
            self.__setattr__("b_" + class_name, loss.__getattr__("Loss_" + class_name).b)
            self.__setattr__("alpha_" + class_name, loss.__getattr__("Loss_" + class_name).alpha)

        self.model = model
        
        if not args.device:
            self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        else:
            self.device = args.device
        
        self.lr = lr
        self.gamma = gamma
        self.clip_value = clip_value
        self.weight_decay = self.args.weight_decay
            
        # TODO! 
        self.model_ref = self.init_model_ref()
        self.model_acc = self.init_model_acc()
        
        self.T = 0
        self.step_counts = 0
    
        def get_parameters(params):
            for p in params:
                yield p
        
        self.params = get_parameters(
            list(model.parameters()) + 
            [self.__getattribute__("a_" + self.args.ID_TO_LABEL[class_idx]) for class_idx in range(self.args.num_class) if (self.args.ID_TO_LABEL[class_idx].startswith(self.label))] +
            [self.__getattribute__("b_" + self.args.ID_TO_LABEL[class_idx]) for class_idx in range(self.args.num_class) if (self.args.ID_TO_LABEL[class_idx].startswith(self.label))] 
        )
        
        self.defaults = dict(
            lr=self.lr, 
            gamma=gamma, 
            clip_value=self.clip_value,
            weight_decay=self.weight_decay,
            model_ref = self.model_ref,
            model_acc = self.model_acc
        )
        
        for class_idx in range(self.args.num_class):
            class_name = self.args.ID_TO_LABEL[class_idx]
            if not class_name.startswith(self.label):
                continue
            self.defaults["p_" + class_name] = self.__getattribute__("p_" + class_name)
            self.defaults["a_" + class_name] = self.__getattribute__("a_" + class_name)
            self.defaults["b_" + class_name] = self.__getattribute__("b_" + class_name)
            self.defaults["margin_" + class_name] = self.__getattribute__("margin_" + class_name)
            self.defaults["alpha_" + class_name] = self.__getattribute__("alpha_" + class_name)

        super(OneVsAllPESG, self).__init__(self.params, self.defaults)
     
    def init_model_ref(self):
         self.model_ref = []
         for var in list(self.model.parameters()) + [self.__getattribute__("a_" + self.args.ID_TO_LABEL[class_idx]) for class_idx in range(self.args.num_class) if (self.args.ID_TO_LABEL[class_idx].startswith(self.label))] + [self.__getattribute__("b_" + self.args.ID_TO_LABEL[class_idx]) for class_idx in range(self.args.num_class) if (self.args.ID_TO_LABEL[class_idx].startswith(self.label))]:
            self.model_ref.append(torch.empty(var.shape).normal_(mean=0, std=0.01).to(self.device))
         return self.model_ref
     
    def init_model_acc(self):
        self.model_acc = []
        for var in list(self.model.parameters()) + [self.__getattribute__("a_" + self.args.ID_TO_LABEL[class_idx]) for class_idx in range(self.args.num_class) if (self.args.ID_TO_LABEL[class_idx].startswith(self.label))] + [self.__getattribute__("b_" + self.args.ID_TO_LABEL[class_idx]) for class_idx in range(self.args.num_class) if (self.args.ID_TO_LABEL[class_idx].startswith(self.label))]:
            self.model_acc.append(torch.zeros(var.shape, dtype=torch.float32,  device=self.device, requires_grad=False).to(self.device)) 
        return self.model_acc
    
    @property    
    def optim_steps(self):
        return self.step_counts
    
    @property
    def get_params(self):
        return list(self.model.parameters())
    
    def update_lr(self, lr):
        self.param_groups[0]['lr']=lr

    @torch.no_grad()
    def step(self):
        """Performs a single optimization step.
        """
        for group in self.param_groups:
            weight_decay = group['weight_decay']
            clip_value = group['clip_value']
            self.lr =  group['lr']
            
            gamma = group['gamma']
           
            model_ref = group['model_ref']
            model_acc = group['model_acc']

            # updates
            for i, p in enumerate(group['params']):
                if p.grad is None:
                    continue
                p.data = p.data - group['lr']*(torch.clamp(p.grad.data , -clip_value, clip_value) + 1/gamma*(p.data - model_ref[i].data) + weight_decay*p.data)
                model_acc[i].data = model_acc[i].data + p.data

            for class_idx in range(self.args.num_class):
                class_name = self.args.ID_TO_LABEL[class_idx]
                if not class_name.startswith(self.label):
                    continue
                group['alpha_' + class_name].data = group['alpha_' + class_name].data + group['lr']*(
                    2*(group['margin_' + class_name] + group['b_' + class_name].data - group['a_' + class_name].data) -
                    2*group['alpha_' + class_name].data
                )
                group['alpha_' + class_name].data = torch.clamp(group['alpha_' + class_name].data, 0, 999)

        self.T += 1  
        self.step_counts += 1

    def zero_grad(self):
        self.model.zero_grad()
        for class_idx in range(self.args.num_class):
            class_name = self.args.ID_TO_LABEL[class_idx]
            if not class_name.startswith(self.label):
                continue
            self.__getattribute__("a_" + class_name).grad = None
            self.__getattribute__("b_" + class_name).grad = None
            self.__getattribute__("alpha_" + class_name).grad = None
        
    def update_regularizer(self, decay_factor=None):
        if decay_factor != None:
            self.param_groups[0]['lr'] = self.param_groups[0]['lr']/decay_factor
            print ('Reducing learning rate to %.5f @ T=%s!'%(self.param_groups[0]['lr'], self.T))
        print ('Updating regularizer @ T=%s!'%(self.T))
        for i, param in enumerate(self.model_ref):
            self.model_ref[i].data = self.model_acc[i].data/self.T
        for i, param in enumerate(self.model_acc):
            self.model_acc[i].data = torch.zeros(param.shape, dtype=torch.float32, device=self.device,  requires_grad=False).to(self.device)
        self.T = 0