import truecase
import re
import math
import random
from itertools import chain
import numpy as np
import os

LABEL_TO_en3IDs = {'O': 0, 'B-ENTITY': 2, 'I-ENTITY': 1}
en3ID_TO_LABEL = {value: key for key, value in LABEL_TO_en3IDs.items()}

DICT_LABELS = {
    "conll": {
        "LABEL_TO_ID": {
            'O': 0, 'B-MISC': 1, 'I-MISC': 2, 'B-PER': 3, 'I-PER': 4,
            'B-ORG': 5, 'I-ORG': 6, 'B-LOC': 7, 'I-LOC': 8
        },
        "LABEL_TO_en3ID": {
            'O': 0, 'B-MISC': 2, 'I-MISC': 1, 'B-PER': 2, 'I-PER': 1,
            'B-ORG': 2, 'I-ORG': 1, 'B-LOC': 2, 'I-LOC': 1
        },
        "LABEL_TO_en2ID": {
            'O': 0, 'B-MISC': 1, 'I-MISC': 1, 'B-PER': 1, 'I-PER': 1,
            'B-ORG': 1, 'I-ORG': 1, 'B-LOC': 1, 'I-LOC': 1
        },
        "LABEL_TO_be2ID": {
            'O': 0, 'B-MISC': 1, 'I-MISC': 0, 'B-PER': 1, 'I-PER': 0,
            'B-ORG': 1, 'I-ORG': 0, 'B-LOC': 1, 'I-LOC': 0
        }
    },
    "bio": {
        "LABEL_TO_ID": {'O': 0, 'B-ENTITY': 1, 'I-ENTITY': 2},
        "LABEL_TO_en3ID": {'O': 0, 'B-ENTITY': 2, 'I-ENTITY': 1},
        "LABEL_TO_en2ID": {'O': 0, 'B-ENTITY': 1, 'I-ENTITY': 1},
        "LABEL_TO_be2ID": {'O': 0, 'B-ENTITY': 1, 'I-ENTITY': 0}
    },
    "ontonotes5": {
        "LABEL_TO_ID": {'O': 0, 'B-CARDINAL': 1, 'B-PERSON': 2, 'I-PERSON': 3, 'B-GPE': 4, 'I-GPE': 5,
                    'B-LOC': 6, 'B-NORP': 7, 'I-DATE': 8, 'B-DATE': 9, 'B-EVENT': 10, 'I-EVENT': 11, 'B-ORG': 12,
                    'I-ORG': 13, 'B-PRODUCT': 14, 'B-FAC': 15, 'B-ORDINAL': 16, 'B-TIME': 17, 'I-TIME': 18, 'I-NORP': 19,
                    'I-PRODUCT': 20, 'I-FAC': 21, 'I-CARDINAL': 22, 'B-MONEY': 23, 'I-MONEY': 24, 'I-LOC': 25,
                    'B-PERCENT': 26, 'I-PERCENT': 27, 'B-WORK_OF_ART': 28, 'I-WORK_OF_ART': 29, 'B-QUANTITY': 30,
                    'I-QUANTITY': 31, 'B-LAW': 32, 'B-LANGUAGE': 33, 'I-LAW': 34, 'I-ORDINAL': 35, 'I-LANGUAGE': 36},
        "LABEL_TO_en3ID": {'O': 0, 'B-CARDINAL': 2, 'B-PERSON': 2, 'I-PERSON': 1, 'B-GPE': 2, 'I-GPE': 1,
                    'B-LOC': 2, 'B-NORP': 2, 'I-DATE': 1, 'B-DATE': 2, 'B-EVENT': 2, 'I-EVENT': 1, 'B-ORG': 2,
                    'I-ORG': 1, 'B-PRODUCT': 2, 'B-FAC': 2, 'B-ORDINAL': 2, 'B-TIME': 2, 'I-TIME': 1, 'I-NORP': 1,
                    'I-PRODUCT': 1, 'I-FAC': 1, 'I-CARDINAL': 1, 'B-MONEY': 2, 'I-MONEY': 1, 'I-LOC': 1,
                    'B-PERCENT': 2, 'I-PERCENT': 1, 'B-WORK_OF_ART': 2, 'I-WORK_OF_ART': 1, 'B-QUANTITY': 2,
                    'I-QUANTITY': 1, 'B-LAW': 2, 'B-LANGUAGE': 2, 'I-LAW': 1, 'I-ORDINAL': 1, 'I-LANGUAGE': 1},
        "LABEL_TO_en2ID": {'O': 0, 'B-CARDINAL': 1, 'B-PERSON': 1, 'I-PERSON': 1, 'B-GPE': 1, 'I-GPE': 1,
                    'B-LOC': 1, 'B-NORP': 1, 'I-DATE': 1, 'B-DATE': 1, 'B-EVENT': 1, 'I-EVENT': 1, 'B-ORG': 1,
                    'I-ORG': 1, 'B-PRODUCT': 1, 'B-FAC': 1, 'B-ORDINAL': 1, 'B-TIME': 1, 'I-TIME': 1, 'I-NORP': 1,
                    'I-PRODUCT': 1, 'I-FAC': 1, 'I-CARDINAL': 1, 'B-MONEY': 1, 'I-MONEY': 1, 'I-LOC': 1,
                    'B-PERCENT': 1, 'I-PERCENT': 1, 'B-WORK_OF_ART': 1, 'I-WORK_OF_ART': 1, 'B-QUANTITY': 1,
                    'I-QUANTITY': 1, 'B-LAW': 1, 'B-LANGUAGE': 1, 'I-LAW': 1, 'I-ORDINAL': 1, 'I-LANGUAGE': 1},
        "LABEL_TO_be2ID": {'O': 0, 'B-CARDINAL': 1, 'B-PERSON': 1, 'I-PERSON': 0, 'B-GPE': 1, 'I-GPE': 0,
                    'B-LOC': 1, 'B-NORP': 1, 'I-DATE': 0, 'B-DATE': 1, 'B-EVENT': 1, 'I-EVENT': 0, 'B-ORG': 1,
                    'I-ORG': 0, 'B-PRODUCT': 1, 'B-FAC': 1, 'B-ORDINAL': 1, 'B-TIME': 1, 'I-TIME': 0, 'I-NORP': 0,
                    'I-PRODUCT': 0, 'I-FAC': 0, 'I-CARDINAL': 0, 'B-MONEY': 1, 'I-MONEY': 0, 'I-LOC': 0,
                    'B-PERCENT': 1, 'I-PERCENT': 0, 'B-WORK_OF_ART': 1, 'I-WORK_OF_ART': 0, 'B-QUANTITY': 1,
                    'I-QUANTITY': 0, 'B-LAW': 1, 'B-LANGUAGE': 1, 'I-LAW': 0, 'I-ORDINAL': 0, 'I-LANGUAGE': 0}
    }
}


def true_case(tokens):
    word_lst = [(w, idx) for idx, w in enumerate(
        tokens) if all(c.isalpha() for c in w)]
    lst = [w for w, _ in word_lst if re.match(r'\b[A-Z\.\-]+\b', w)]

    if len(lst) and len(lst) == len(word_lst):
        parts = truecase.get_true_case(' '.join(lst)).split()
        if len(parts) != len(word_lst):
            return tokens
        for (w, idx), nw in zip(word_lst, parts):
            tokens[idx] = nw
    return tokens


def process_instance(crfopt, data_format, words, labels, tokenizer, num_class, max_seq_length):
    tokens, token_labels, token_labels_en3, token_labels_en2, token_labels_be2 = [], [], [], [], []
    for word, label in zip(words, labels):
        if crfopt:
            tokens += [word]
            token_label = [DICT_LABELS[data_format]["LABEL_TO_ID"][label]]
            token_labels += token_label
            token_label_en3 = [DICT_LABELS[data_format]["LABEL_TO_en3ID"][label]]
            token_labels_en3 += token_label_en3
            token_label_en2 = [DICT_LABELS[data_format]["LABEL_TO_en2ID"][label]]
            token_labels_en2 += token_label_en2
            token_label_be2 = [DICT_LABELS[data_format]["LABEL_TO_be2ID"][label]]
            token_labels_be2 += token_label_be2
        else:
            tokenized = tokenizer.tokenize(word)
            tokens += tokenized
            token_label = [DICT_LABELS[data_format]["LABEL_TO_ID"][label]] + [-1] * (len(tokenized) - 1)
            token_labels += token_label
            token_label_en3 = [DICT_LABELS[data_format]["LABEL_TO_en3ID"][label]] + [-1] * (len(tokenized) - 1)
            token_labels_en3 += token_label_en3
            token_label_en2 = [DICT_LABELS[data_format]["LABEL_TO_en2ID"][label]] + [-1] * (len(tokenized) - 1)
            token_labels_en2 += token_label_en2
            token_label_be2 = [DICT_LABELS[data_format]["LABEL_TO_be2ID"][label]] + [-1] * (len(tokenized) - 1)
            token_labels_be2 += token_label_be2
    
    assert len(tokens) == len(token_labels)
    
    tokens, token_labels, token_labels_en3, token_labels_en2, token_labels_be2 = \
        tokens[:max_seq_length - 2], token_labels[:max_seq_length - 2], token_labels_en3[:max_seq_length - 2], \
        token_labels_en2[:max_seq_length - 2], token_labels_be2[:max_seq_length - 2]
    
    input_ids = tokenizer.convert_tokens_to_ids(tokens)
    input_ids = tokenizer.build_inputs_with_special_tokens(input_ids)
    token_labels = [-1] + token_labels + [-1]
    token_labels_en3 = [-1] + token_labels_en3 + [-1]
    token_labels_en2 = [-1] + token_labels_en2 + [-1]
    token_labels_be2 = [-1] + token_labels_be2 + [-1]

    if (num_class == 3):
        token_labels = token_labels_en3

    return {
        "input_ids": input_ids,
        "labels": token_labels,
        "labels_en3": token_labels_en3,
        "labels_en2": token_labels_en2,
        "labels_be2": token_labels_be2,
    }


def read_conll(args, file_in, tokenizer, num_class, max_seq_length=512,
               num_training_sentences=None, part=None,
               imbalance_generator=False, generate_imbRatio=None, generate_imbRatio_error=None, randMaxSeeds=None):
    words, labels = [], []
    all_words, all_labels = [], []
    examples = []
    is_title = False
    curr_num_sentences = 0
    num_en = 0
    num_be = 0
    num_tokens = 0

    data_format =  args.corpus if args.corpus in ["conll", "ontonotes5"] else "bio"

    seed = False
    with open(file_in, "r") as fh:
        for line in fh:
            line = line.strip()
            if line.startswith("-DOCSTART-"):
                is_title = True
                continue
            if len(line) > 0:
                line = line.split()
                word = line[0]
                label = line[-1]
                words.append(word)
                labels.append(label)
            else:
                if len(words) > 0:
                    if is_title:
                        words = true_case(words)
                        is_title = False
                    assert len(words) == len(labels)
                    curr_num_sentences += 1

                    flag = False

                    if num_training_sentences and part:  # if both of these happen
                        if curr_num_sentences in range(int(num_training_sentences*(part-1)), int(num_training_sentences*part)):
                            flag = True
                    else:
                        flag = True

                    if flag:
                        proc_instance = process_instance(
                            args.crfopt,
                            data_format,
                            words, labels, tokenizer, num_class, max_seq_length
                        )
                        num_tokens += len(proc_instance["labels_en2"])
                        num_en += sum([1 if i == 1 else 0 for i in proc_instance["labels_en2"]])
                        num_be += sum([1 if i == 1 else 0 for i in proc_instance["labels_be2"]])
                        # num_b += sum([1 if i == 1 else 0 for i in proc_instance["labels"]])

                        all_words.append(words)
                        all_labels.append(labels)

                        if imbalance_generator:
                            pass
                        else:
                            examples.append(proc_instance)

                    if part:
                        if num_training_sentences and curr_num_sentences == num_training_sentences*part:
                            break
                    else:
                        if imbalance_generator:
                            pass
                        else:
                            if num_training_sentences and curr_num_sentences == num_training_sentences:
                                break

                    words, labels = [], []

    if imbalance_generator:
        mix_sent_idx, en_ratio, be_ratio, seed = generate_by_sentRatio(
            data_format,
            all_labels, num_training_sentences, generate_imbRatio, generate_imbRatio_error, randMaxSeeds
        )

        assert mix_sent_idx is not None

        for eachSentIdx in mix_sent_idx:
            words = all_words[eachSentIdx]
            labels = all_labels[eachSentIdx]
            examples.append(
                process_instance(
                    args.crfopt,
                    data_format,
                    words, labels, tokenizer, num_class, max_seq_length
                )
            )

        print("en_ratio:", en_ratio, "be_ratio:", be_ratio)

    if file_in == os.path.join(args.data_dir, "train.txt"):
        if args.part and args.num_training_sentences:
            train_features_subset = examples[int(args.num_training_sentences*(args.part-1)):int(args.num_training_sentences*args.part)]
        elif args.num_training_sentences:
            train_features_subset = examples[:int(args.num_training_sentences)]

        if args.num_training_sentences:
            try:
                assert len(train_features_subset) == args.num_training_sentences
                examples = train_features_subset
            except:
                random.seed(args.part)
                examples = random.sample(examples, k=args.num_training_sentences)
    else:
        pass
    
    return examples, seed


def generate_by_sentRatio(data_format, all_labels, num_sent, imbRatio, error, randMaxSeeds):
    tokenidx = list(range(len(all_labels)))
    enidx = list(filter(lambda x: True if (1 in [DICT_LABELS[data_format]["LABEL_TO_en2ID"][y] for y in all_labels[x]]) else False, tokenidx))
    noEnidx = list(set(tokenidx) - set(enidx))

    for seed in range(1, randMaxSeeds):
        num_en = math.ceil(num_sent * imbRatio)

        while num_en < num_sent:
            num_noEn = num_sent - num_en
            li_sent = generate_imbalance(num_noEn, num_en, noEnidx, enidx)
            en_ratio, be_ratio = get_imbalance(data_format, all_labels, li_sent)
            
            if en_ratio < imbRatio * (1 + error) and en_ratio > imbRatio * (1 - error):
                return li_sent, en_ratio, be_ratio, seed
            
            num_en = num_en + 1


def generate_imbalance(num_noEn, num_en, noEnidx, enidx):
    li_noEn = random.choices(noEnidx, k=num_noEn)
    li_en = random.choices(enidx, k=num_en)

    return sorted(li_noEn + li_en)


def get_imbalance(data_format, all_labels, li_mix):

    mix_labels = list(chain.from_iterable([all_labels[i] for i in li_mix]))

    mix_en2_labels = np.array([DICT_LABELS[data_format]["LABEL_TO_en2ID"][x] for x in mix_labels])
    mix_be2_labels = np.array([DICT_LABELS[data_format]["LABEL_TO_be2ID"][x] for x in mix_labels])

    en_ratio = mix_en2_labels.sum()/len(mix_en2_labels)
    be_ratio = mix_be2_labels.sum()/len(mix_be2_labels)

    return en_ratio, be_ratio