import argparse, os, time, math, pdb, gc, time, random, copy
from itertools import chain
from tqdm import tqdm
import numpy as np
import pandas as pd

import torch
import torch.nn as nn
from torch.utils.data import DataLoader
from transformers import AutoTokenizer
from torch.optim import Adam, AdamW
from transformers.optimization import get_linear_schedule_with_warmup
from torch.cuda.amp import autocast, GradScaler

import seqeval.metrics
import wandb

from model import *
from utils import *
from losses import *
from optimizers import *
from prepro import read_conll, DICT_LABELS, en3ID_TO_LABEL


def getEvalResults(args, benchmarks, model, num_steps, bestlogits, tokenizer, best_f1, best_output, best_dev_output, Loss=None):
    for tag, features in benchmarks:
        if (tag == 'dev'):
            results, report, tempdevlogits = evaluate(args, model, features, tag=tag)
            if not args.wandboff:
                wandb.log(results, step=num_steps)
            cur_f1_score = results["dev_f1"]
            cur_dev_output = results
        
        elif (tag == 'test'):
            results, report, temptestlogits = evaluate(args, model, features, tag=tag)
            if not args.wandboff:
                wandb.log(results, step=num_steps)
            cur_output = results
    
    if cur_f1_score > best_f1:
        best_f1 = cur_f1_score
        bestlogits['dev'] = tempdevlogits
        bestlogits['test'] = temptestlogits
        best_output = cur_output
        best_dev_output = cur_dev_output
        tokenizer.save_pretrained(args.output_dir)

    return best_f1, best_output, best_dev_output, bestlogits


def evaluate(args, model, features, tag):
    dataloader = DataLoader(features, batch_size=128, shuffle=False, collate_fn=collate_fn, drop_last=False)
    preds, keys, bestlogits = [], [], []

    for batch in dataloader:
        model.eval()
        batch = {key: value.to(args.device) for key, value in batch.items()}
        keys += batch['labels'].cpu().numpy().flatten().tolist()
        mask = ~torch.eq(batch["labels"], -1)
        batch['labels'] = None
        
        with torch.no_grad():
            if args.meta_learning:
                preds012 = [None]*args.num_class
                logits, labels = model(batch)
                
                for i in range(args.num_class):
                    class_name = args.ID_TO_LABEL[i]
                    preds012[i] = torch.sigmoid(logits[class_name]).view(-1)
                
                stacked_logits = torch.stack(preds012, -1)
                bestlogits.append(stacked_logits.cpu().detach())
                preds += torch.argmax(stacked_logits, dim=-1).cpu().detach().numpy().flatten().tolist()
            
            else:
                logits = model(batch)[0]
                
                if args.crfopt:
                    temppreds = list(chain.from_iterable(model.crf.decode(logits[:, 1:], mask=mask[:, 1:].bool())))
                else:
                    temppreds = np.argmax(logits.cpu().numpy(), axis=-1).tolist()
                    bestlogits.append(logits.cpu().detach())
                
                preds += temppreds

    if (args.crfopt):
        keys = [key for key in keys if key != -1]
        bestlogits = preds
    
    preds, keys = list(zip(*[[pred, key] for pred, key in zip(preds, keys) if key != -1]))

    preds = [args.ID_TO_LABEL[pred] for pred in preds]
    keys = [args.ID_TO_LABEL[key] for key in keys]

    model.zero_grad()

    f1 = seqeval.metrics.f1_score([keys], [preds], scheme="IOB2")  # IOB2 works for BIO
    precision = seqeval.metrics.precision_score([keys], [preds], scheme="IOB2")
    recall = seqeval.metrics.recall_score([keys], [preds], scheme="IOB2")
    result = seqeval.metrics.classification_report([keys], [preds], digits=4, scheme="IOB2")

    output = {
        tag + "_f1": f1,
        tag + "_precision": precision,
        tag + "_recall": recall,
    }

    return output, result, bestlogits


def train(args, model, train_features, benchmarks, tokenizer):
    logger = get_logger(args)
    model.to(args.device)
    
    if args.n_gpu > 1:
        model = torch.nn.DataParallel(model)

    train_dataloader = DataLoader(train_features, batch_size=args.batch_size, shuffle=True, collate_fn=collate_fn, drop_last=True)
    total_steps = int(len(train_dataloader) * args.num_train_epochs // args.gradient_accumulation_steps)
    warmup_steps = int(total_steps * args.warmup_ratio)

    if (args.crfopt):
        no_decay = ["bias", "LayerNorm.weight"]
        
        bert_param_optimizer = list(model.bertfamily.named_parameters())
        crf_param_optimizer = list(model.crf.named_parameters())
        linear_param_optimizer = list(model.classifier.named_parameters())
        
        optimizer_grouped_parameters = [
            {'params': [p for n, p in bert_param_optimizer if not any(nd in n for nd in no_decay)],
             'weight_decay': args.weight_decay, 'lr': args.learning_rate},
            {'params': [p for n, p in bert_param_optimizer if any(nd in n for nd in no_decay)], 'weight_decay': 0.0,
             'lr': args.learning_rate},

            {'params': [p for n, p in crf_param_optimizer if not any(nd in n for nd in no_decay)],
             'weight_decay': args.weight_decay, 'lr': args.learning_rate},
            {'params': [p for n, p in crf_param_optimizer if any(nd in n for nd in no_decay)], 'weight_decay': 0.0,
             'lr': args.learning_rate},

            {'params': [p for n, p in linear_param_optimizer if not any(nd in n for nd in no_decay)],
             'weight_decay': args.weight_decay, 'lr': args.learning_rate},
            {'params': [p for n, p in linear_param_optimizer if any(nd in n for nd in no_decay)],
             'weight_decay': 0.0,
             'lr': args.learning_rate}
        ]
        
        optimizer = AdamW(optimizer_grouped_parameters, lr=args.learning_rate, eps=args.eps)
        scheduler = get_linear_schedule_with_warmup(optimizer, num_warmup_steps=warmup_steps, num_training_steps=total_steps)
    
    else:
        optimizer = Adam(model.parameters(), lr=args.learning_rate, eps=args.eps)
        scheduler = get_linear_schedule_with_warmup(optimizer, num_warmup_steps=warmup_steps, num_training_steps=total_steps)
    
    scaler = GradScaler()

    num_steps = 0
    best_f1 = -1.0
    bestlogits = {}
    best_output = None
    best_dev_output = None

    for epoch in range(int(args.num_train_epochs)):
        for step, batch in enumerate(tqdm(train_dataloader)):
            if epoch==0 and step==0:
                best_f1, best_output, best_dev_output, bestlogits = getEvalResults(
                    args, benchmarks, model, num_steps, bestlogits, tokenizer, best_f1, best_output, best_dev_output
                )
                
            model.train()

            if num_steps < args.alpha_warmup_ratio * total_steps:
                args.alpha_t = 0.0
            else:
                args.alpha_t = args.alpha

            batch = {key: value.to(args.device) for key, value in batch.items()}
            with autocast():
                outputs = model(batch)

            if args.debug:
                pdb.set_trace()

            loss = outputs[0]

            if args.n_gpu > 1:
                loss = loss.mean()  # mean() to average on multi-gpu.
            if args.gradient_accumulation_steps > 1:
                loss = loss / args.gradient_accumulation_steps

            scaler.scale(loss).backward()
            if step % args.gradient_accumulation_steps == 0:
                num_steps += 1
                scale = scaler.get_scale()
                scaler.unscale_(optimizer)
                torch.nn.utils.clip_grad_norm_(model.parameters(), args.max_grad_norm)
                scaler.step(optimizer)
                scaler.update()
                skip_lr_sched = (scale > scaler.get_scale())
                if not skip_lr_sched:
                    scheduler.step()
                model.zero_grad()
                if not args.wandboff:
                    wandb.log({'loss': loss.item()}, step=num_steps)

            if step == len(train_dataloader) - 1:
                best_f1, best_output, best_dev_output, bestlogits = getEvalResults(
                    args, benchmarks, model, num_steps, bestlogits, tokenizer, best_f1, best_output, best_dev_output
                )

    saveBestLogits(args, bestlogits)

    return best_output, best_dev_output, bestlogits


def trainM(args, model, train_features, benchmarks, tokenizer):
    logger = get_logger(args)
    model.to(args.device)

    train_dataloader = DataLoader(train_features, batch_size=args.batch_size, shuffle=True, collate_fn=collate_fn, drop_last=True)
    total_steps = int(len(train_dataloader) *args.num_train_epochs // args.gradient_accumulation_steps)
    
    if args.comopt:
        total_steps = total_steps//2
    
    warmup_steps = int(total_steps * args.warmup_ratio)

    Loss = MAUCNERLoss(args)    

    optimizer = MPESG(model, Loss, args, args.learning_rate_auc)
    scheduler = get_linear_schedule_with_warmup(
        optimizer, num_warmup_steps=warmup_steps, 
        num_training_steps=total_steps*1.05
    )
    
    if (args.comopt):
        adamoptimizer = Adam(model.parameters(), lr=args.learning_rate, eps=args.eps)
        adamscheduler = get_linear_schedule_with_warmup(
            adamoptimizer, num_warmup_steps=warmup_steps, 
            num_training_steps=total_steps
        )
        Loss.backend="ce"

    scaler = GradScaler()

    num_steps = 0
    best_f1 = -1.0
    best_output = None
    best_dev_output = None
    cross_entropy_loss_fnt = nn.CrossEntropyLoss(ignore_index=-1)
    lost_list = []
    bestlogits= {}

    for epoch in range(int(args.num_train_epochs)):
        for step, batch in enumerate(tqdm(train_dataloader)):
            model.train()
            if num_steps < args.alpha_warmup_ratio * total_steps:
                args.alpha_t = 0.0
            else:
                args.alpha_t = args.alpha

            batch = {key: value.to(args.device) for key, value in batch.items()}
            
            with autocast():
                logits, labels = model(batch)

            if args.debug:
                pdb.set_trace()

            comopt_and_ce_backend = (Loss.backend == "ce" if args.comopt else False)

            if comopt_and_ce_backend:
                logits_ce = [logits[args.ID_TO_LABEL[i]] for i in range(args.num_class)]
                loss = cross_entropy_loss_fnt(torch.stack(logits_ce, dim=1).squeeze(), labels["labels"].view(-1))
                Loss.backend = "auc"
            
            else:
                loss, loss_dict = Loss(
                    logits, labels["labels"].view(-1), 
                    random.sample(list(args.LABEL_TO_ID.keys()), math.ceil(args.num_class/args.task_division)) if args.meta_learning else None
                )
                lost_list.append((loss.item(),loss_dict))
                Loss.backend="ce"

            if args.gradient_accumulation_steps > 1:
                loss = loss / args.gradient_accumulation_steps

            scaler.scale(loss).backward()

            if step % args.gradient_accumulation_steps == 0:
                num_steps += 1
                if (comopt_and_ce_backend):
                    scale = scaler.get_scale()
                    scaler.unscale_(adamoptimizer)
                    torch.nn.utils.clip_grad_norm_(model.parameters(), args.max_grad_norm)
                    scaler.step(adamoptimizer)
                    scaler.update()
                    skip_lr_sched = (scale > scaler.get_scale())
                    
                    if not skip_lr_sched and args.learning_schedule:
                        adamscheduler.step()
                    
                    adamoptimizer.zero_grad()

                else:
                    scale = scaler.get_scale()
                    scaler.unscale_(optimizer)
                    torch.nn.utils.clip_grad_norm_(model.parameters(), args.max_grad_norm)
                    scaler.step(optimizer)
                    scaler.update()
                    skip_lr_sched = (scale > scaler.get_scale())
                    
                    if not skip_lr_sched and args.learning_schedule:
                        scheduler.step()
                    
                    optimizer.zero_grad()
        
        if epoch > 0:
            if not args.wandboff:
                wandb.log({'train_loss': lost_list[-1][0]}, step=epoch)
                for key, value in lost_list[-1][1].items():
                    wandb.log({key: value}, step=epoch)
        
        best_f1, best_output, best_dev_output, bestlogits = getEvalResults(
            args, benchmarks, model, epoch, bestlogits, tokenizer, best_f1, best_output, best_dev_output, Loss
        )

    saveBestLogits(args, bestlogits)

    return best_output, best_dev_output, Loss, bestlogits


def evaluateOVA(args, model, features, classbestlogits, label, best_result, tag, stepflag, epoch):
    dataloader = DataLoader(features, batch_size=128, shuffle=False, collate_fn=collate_fn, drop_last=False)

    preds, keys = [], []
    preds_label = {}
    bestlogits = {}
    preds_label = [[] for i in range(args.num_class)]
    bestlogits = [[] for i in range(args.num_class)]

    for batch_idx, batch in enumerate(dataloader):
        model.eval()

        batch = {key: value.to(args.device) for key, value in batch.items()}
        keys += batch['labels'].cpu().numpy().flatten().tolist()
        batch['labels'] = None

        with torch.no_grad():
            logits, labels = model(batch)
            if (label=="O"):
                for class_idx in range(args.num_class):
                    class_name = args.ID_TO_LABEL[class_idx]
                    preds_label[class_idx] += torch.sigmoid(logits[class_name]).view(-1).cpu().detach().numpy().flatten().tolist()
                    bestlogits[class_idx] += torch.sigmoid(logits[class_name]).cpu().detach().numpy().flatten().tolist()
            
            elif (label=="I"):
                for class_idx in range(args.num_class):
                    class_name = args.ID_TO_LABEL[class_idx]
            
                    if class_name=="O":
                        preds_label[class_idx] = classbestlogits["O"][tag + "_" + "logits"][class_idx]
                    else:
                        preds_label[class_idx] += torch.sigmoid(logits[class_name]).view(-1).cpu().detach().numpy().flatten().tolist()
                        bestlogits[class_idx] += torch.sigmoid(logits[class_name]).cpu().detach().numpy().flatten().tolist()

            elif (label=="B"):
                for class_idx in range(args.num_class):
                    class_name = args.ID_TO_LABEL[class_idx]
                    
                    if class_name=="O": 
                        preds_label[class_idx] = classbestlogits["O"][tag + "_" + "logits"][class_idx]
                    elif class_name.startswith("I"):
                        preds_label[class_idx] = classbestlogits["I"][tag + "_" + "logits"][class_idx]
                    else:
                        preds_label[class_idx] += torch.sigmoid(logits[class_name]).view(-1).cpu().detach().numpy().flatten().tolist()
                        bestlogits[class_idx] += torch.sigmoid(logits[class_name]).cpu().detach().numpy().flatten().tolist()
                
    preds_label_tensors = [torch.tensor(curr_preds) for curr_preds in preds_label]

    preds012 = torch.stack(preds_label_tensors, -1)
    preds = torch.argmax(preds012, dim=-1).cpu().detach().numpy().flatten().tolist()

    preds, keys = list(zip(*[[pred, key] for pred, key in zip(preds, keys) if key != -1]))

    preds = [args.ID_TO_LABEL[pred] for pred in preds]
    keys = [args.ID_TO_LABEL[key] for key in keys]

    f1 = seqeval.metrics.f1_score([keys], [preds], scheme="IOB2")  # IOB2 works for BIO
    precision = seqeval.metrics.precision_score([keys], [preds], scheme="IOB2")
    recall = seqeval.metrics.recall_score([keys], [preds], scheme="IOB2")

    result = seqeval.metrics.classification_report([keys], [preds], digits=4, scheme="IOB2")

    output = {
        tag + "_f1": f1,
        tag + "_precision": precision,
        tag + "_recall": recall
    }

    return output, result, bestlogits
  

def trainOVA(args, model, train_features, benchmarks, tokenizer):
    model.to(args.device)

    train_dataloader = DataLoader(train_features, batch_size=args.batch_size, shuffle=True, collate_fn=collate_fn, drop_last=True)
    total_steps = int(len(train_dataloader) *args.num_train_epochs // args.gradient_accumulation_steps)
    
    if args.comopt:
        total_steps = total_steps//2
    
    warmup_steps = int(total_steps * args.warmup_ratio)
    
    label_set = ["O", "I",  "B"]
    best_score = {"B":-1.0, "I":-1.0, "O":-1.0}

    Loss_Dict = {}
    Optimizer_Dict = {}
    Scheduler_Dict = {}
    AdamOptimizer_Dict = {}
    AdamScheduler_Dict = {}

    for label in label_set:
        Loss_Dict['Loss_' + label] = OneVsAllAUCNERLoss(args, label)
        Optimizer_Dict['Optimizer_' + label] = OneVsAllPESG(
            model, Loss_Dict['Loss_' + label],
            args, label, lr=args.learning_rate_auc
        )
    
        Scheduler_Dict['Scheduler_' + label] = get_linear_schedule_with_warmup(
            Optimizer_Dict['Optimizer_' + label], num_warmup_steps=warmup_steps, num_training_steps= total_steps
        )   
    
        if args.comopt or args.ceopt:
            AdamOptimizer_Dict['AdamOptimizer_' + label] = Adam(model.parameters(), lr=args.learning_rate, eps=args.eps)
            AdamScheduler_Dict['AdamScheduler_' + label] = get_linear_schedule_with_warmup(
                AdamOptimizer_Dict['AdamOptimizer_' + label], num_warmup_steps=warmup_steps, num_training_steps= total_steps
            )
            Loss_Dict['Loss_' + label].backend="ce"
            if args.ceopt:
                Loss_Dict['Loss_' + label]= OVACEloss(args, label)
        
    if args.ceopt:
        del Optimizer_Dict
        del Scheduler_Dict

    scaler = GradScaler()
    num_steps = 0
    best_output = None
    best_dev_output = None
    loss_dict = {}
    bestlogits= {}
    cross_entropy_loss_fnt = nn.CrossEntropyLoss(ignore_index=-1)

    for epoch in range(int(args.num_train_epochs)):
        for label in label_set:
            for step, batch in enumerate(tqdm(train_dataloader)):
                model.train()

                if num_steps < args.alpha_warmup_ratio * total_steps:
                    args.alpha_t = 0.0
                else:
                    args.alpha_t = args.alpha

                batch = {key: value.to(args.device) for key, value in batch.items()}
                
                with autocast():
                    logits, labels = model(batch)
                
                if args.debug:
                    pdb.set_trace()

                comopt_and_ce_backend = (Loss_Dict['Loss_'+label].backend == "ce" if args.comopt else False)

                if comopt_and_ce_backend:
                    logits_ce = [logits[args.ID_TO_LABEL[i]] for i in range(args.num_class)]
                    loss = cross_entropy_loss_fnt(torch.stack(logits_ce, dim=1).squeeze(), labels["labels"].view(-1))
                    Loss_Dict['Loss_'+label].backend = "auc"
                else:
                    loss, loss_dict = Loss_Dict['Loss_'+label](logits, labels["labels"].view(-1), loss_dict)

                if args.gradient_accumulation_steps > 1:
                    loss = loss / args.gradient_accumulation_steps

                scaler.scale(loss).backward()
                if step % args.gradient_accumulation_steps == 0:
                    num_steps += 1
                    scale = scaler.get_scale()
                    if comopt_and_ce_backend or args.ceopt:
                        scaler.unscale_(AdamOptimizer_Dict['AdamOptimizer_' + label])
                        torch.nn.utils.clip_grad_norm_(model.parameters(), args.max_grad_norm)
                        scaler.step(AdamOptimizer_Dict['AdamOptimizer_' + label])
                        scaler.update()
                        skip_lr_sched = (scale > scaler.get_scale())

                        if not skip_lr_sched and args.learning_schedule:
                            AdamScheduler_Dict['AdamScheduler_' + label].step()
                        
                        AdamOptimizer_Dict['AdamOptimizer_' + label].zero_grad()
                    else:
                        scaler.unscale_(Optimizer_Dict['Optimizer_'+label])
                        torch.nn.utils.clip_grad_norm_(model.parameters(), args.max_grad_norm)
                        scaler.step(Optimizer_Dict['Optimizer_'+label])
                        scaler.update()
                        skip_lr_sched = (scale > scaler.get_scale())
                        
                        if not skip_lr_sched and args.learning_schedule:
                            Scheduler_Dict['Scheduler_'+label].step()
                        
                        Optimizer_Dict['Optimizer_'+label].zero_grad()

                if not comopt_and_ce_backend:
                    if not args.wandboff:
                        wandb.log({'train_loss_' + label : loss.item()}, step=num_steps)
                        for key, value in loss_dict.items():
                            wandb.log({key: value}, step=num_steps)

                stepflag = (step == len(train_dataloader) - 1)
                
                if stepflag:
                    for tag, features in benchmarks:
                        if (tag == 'dev'):
                            results, report, logitstemp_dev = evaluateOVA(args, model, features, bestlogits, label, best_dev_output, tag, stepflag, epoch)
                            if not args.wandboff:
                                wandb.log(results, step=num_steps)
                            cur_score = results["dev_f1"]
                            cur_dev_output = results
                        if (tag == 'test'):
                            results, report, logitstemp_test = evaluateOVA(args, model, features, bestlogits, label, best_output, tag, stepflag, epoch)
                            if not args.wandboff:
                                wandb.log(results, step=num_steps)
                            cur_output = results

                    if cur_score > best_score[label]:
                        best_score[label] = cur_score
                        bestlogits[label] = {"dev_logits": logitstemp_dev, "test_logits": logitstemp_test}
                        best_output = cur_output
                        best_dev_output = cur_dev_output
                        tokenizer.save_pretrained(args.output_dir)

    bestlogits = saveBestLogits(args, bestlogits, label_set)

    return best_output, best_dev_output, Loss_Dict, bestlogits


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument("--project_name", type=str, default="NERAUC", help="name of the project for Wandb to log the results")
    parser.add_argument("--wandboff", default=True, action='store_false', help="Turn on wandb logging")
    parser.add_argument("--data_dir", default="../data", type=str, help="The input data dir. Should contain the .tsv files (or other data files) for the task.")
    parser.add_argument("--results_dir", default="../results", type=str, help="The output directory where the model predictions perfromance will be written.")
    parser.add_argument("--corpus", default="conll", type=str, help="The corpus to be used for training and evaluation.")

    parser.add_argument("--num_class", type=int, default=3, help="Number of classes, 9 for CoNLL 2003, 37 for OntoNotes 5.0, and 3 for NCBI and s800")
    parser.add_argument("--csv_results_name", default=None, type=str, help="The name of the csv file to be saved with the results, if None, the name will be automatically set up")

    parser.add_argument("--model_name_or_path", default="bert-base-cased", type=str, help="Path to pre-trained model")

    parser.add_argument("--max_seq_length", default=128, type=int, help="The maximum total input sequence length after tokenization. Sequences longer than this will be truncated, sequences shorter will be padded.")
    parser.add_argument("--batch_size", default=64, type=int, help="Batch size per GPU/CPU for training.")

    parser.add_argument("--learning_rate", default=1e-5, type=float, help="The initial learning rate for Adam or AdamW optimizer.")
    parser.add_argument("--weight_decay", default=1e-4, type=float, help="Weight decay if we apply some.")
    parser.add_argument("--gradient_accumulation_steps", default=1, type=int, help="Number of updates steps to accumulate before performing a backward/update pass.")
    parser.add_argument("--eps", default=1e-6, type=float, help="Epsilon for Adam optimizer.")
    parser.add_argument("--max_grad_norm", default=1.0, type=float, help="Max gradient norm.")
    parser.add_argument("--warmup_ratio", default=0.1, type=float, help="Linear warmup over warmup_ratio.")
    parser.add_argument("--dropout_prob", default=0.1, type=float, help="Dropout probability.")
    parser.add_argument("--num_train_epochs", default=100.0, type=float, help="Total number of training epochs to perform.")
    parser.add_argument("--seed", type=int, default=42, help="random seed for initialization")
    parser.add_argument("--alpha", type=float, default=50.0)
    parser.add_argument("--alpha_warmup_ratio", default=0.1, type=float)

    parser.add_argument("--output_dir", default="../saveNERmodel", type=str, help="The output directory where the model prediction logits and checkpoints will be written.")
    parser.add_argument("--log_file", default="", type=str, help="The output file where the training and evaluation logs will be written. DEPRECEATED, USED Wandb instead")

    parser.add_argument("--onevsall", default=False, action='store_true', help="Turn on onevsall training")
    parser.add_argument("--meta_learning", default=False, action='store_true', help="Turn on meta learning")
    parser.add_argument("--task_division", default=2.0, type=float, help="Number of tasks to be divided into for Meta Learning, m=num_class/task_division")

    parser.add_argument("--learning_rate_auc", default=1e-1, help="Learning rate for AUC optimizer")
    parser.add_argument("--aucopt", default=False, action='store_true', help="Turn on AUC training, If false, then CE training")
    parser.add_argument("--ceopt", default=False, action='store_true', help="Turn on BCE training, this needs onevsall to be true")
    parser.add_argument("--comopt", default=False, action='store_true', help="Turn on compositional training")
    parser.add_argument("--crfopt", default=False, action='store_true', help="Turn on CRF training")
    parser.add_argument("--bilstmopt", default=False, action='store_true', help="Add BiLSTM layer to CRF training")

    parser.add_argument("--ensembleopt", default=False, action='store_true', help="Turn on ensemble training")
    parser.add_argument("--num_ensembles", default=5, type=int, help="Number of ensembles to be trained")

    parser.add_argument("--optimizertype", default="PESG", type=str, help="Optimizer type")
    parser.add_argument("--learning_schedule", default=True, action='store_false')
    parser.add_argument("--margin", default=1.0, type=float)

    parser.add_argument("--num_training_sentences", default=None, type=int)
    parser.add_argument("--part", default=None, type=int)
    
    parser.add_argument("--imbalance_generator", default=False, action='store_true')
    parser.add_argument("--randMaxSeeds", default=1000000, type=int)
    parser.add_argument("--generate_imbRatio", default=0.01, type=float)
    parser.add_argument("--generate_imbRatio_error", default=0.05, type=float)

    parser.add_argument("--n_gpu", default=None, type=int)
    parser.add_argument("--gpu_no", default="0", type=str)

    parser.add_argument("--debug", default=False, action='store_true', help="Turn on debug mode")
    
    args = parser.parse_args()

    # If the corpus is from biomedical domain, we use biobert
    if args.corpus not in ["conll", "ontonotes5"]:
        args.model_name_or_path = "dmis-lab/biobert-base-cased-v1.1"

    # Setting up the directories to record the results
    args.data_dir = os.path.join(args.data_dir, args.corpus)
    args.output_dir = os.path.join(args.output_dir, args.corpus)
    args.results_dir = os.path.join(args.results_dir, args.corpus)

    if not os.path.exists(args.results_dir):
        os.makedirs(args.results_dir)

    if args.comopt:
        # Change the name of the optimizer in the case of compositional training (alternate between AdamW and PESG)
        args.optimizertype = "PDSCA"
        # Compositional training requires doulbe the normal number of epochs (epochs CE + epochs AUC)
        args.num_train_epochs = args.num_train_epochs*2

    optname = None

    # Naming the learning objective function used for the csv file
    if args.crfopt:
        optname = "CRF"
        if args.bilstmopt:
            optname = "BiLSTM" + optname
    elif args.aucopt:
        optname = "AUC"
        if args.comopt:
            optname = "COMAUC"
    elif args.ceopt:
        optname = "BCE"
    else:
        optname = "CE"

    if args.meta_learning:
        optname = "OVAML" + optname
        args.num_train_epochs = args.num_train_epochs*args.task_division

    if args.onevsall:
        optname = "OVA" + optname

    prefix = "%s%s%s%s%s" % (
        "%s_" % (args.corpus) if args.corpus != "conll" else "",
        optname, 
        "_num_class_%s" % str(args.num_class) if args.num_class != 3 else "",
        '_imb_gen' if args.imbalance_generator else '',
        "_ensemble_%s"%(args.num_ensembles) if args.ensembleopt else "",
    ) 
    
    if args.csv_results_name:
        prefix = str(args.csv_results_name) + "_" + prefix

    results_path = os.path.join(args.results_dir, "%s_results.csv" % (prefix))
    
    if not args.n_gpu:
        args.n_gpu = torch.cuda.device_count()

    while args.batch_size*args.n_gpu > args.num_training_sentences:
        args.batch_size = int(args.batch_size/2)
    
    args.learning_rate = args.learning_rate * np.sqrt(args.batch_size / 64)
    args.learning_rate_auc = args.learning_rate_auc * np.sqrt(args.batch_size / 64)

    # Getting the path to the csv file and Return the number of iterations that you have done for the same settings
    if os.path.exists(results_path):
        csv_file = pd.read_csv(results_path, index_col=False)
        
        if (args.aucopt) or (args.comopt) or (args.ceopt):
            csv_file.learning_rate_auc = csv_file.learning_rate_auc.round(4)
            csv_results = csv_file.loc[
                (csv_file["bert"] == args.model_name_or_path) &
                (csv_file["optimizertype"] == args.optimizertype) &
                (csv_file["num_training_sentences"] == args.num_training_sentences) &
                (csv_file["learning_rate_auc"] == round(args.learning_rate_auc, 4))
            ]

        else:
            csv_results = csv_file.loc[
                (csv_file["bert"] == args.model_name_or_path) &
                (csv_file["num_training_sentences"] == args.num_training_sentences)
            ]

        if args.imbalance_generator:
            csv_results = csv_results.loc[
                (csv_results["generate_imbRatio"] == args.generate_imbRatio) &
                (csv_results["generate_imbRatio_error"] == args.generate_imbRatio_error) &
                (csv_results["randMaxSeeds"] == args.randMaxSeeds) &
                (csv_results["seed"] == args.seed)
            ]
            past_seed = csv_results["seed"].tolist() + [42]

        past_iteration = csv_results.shape[0]
        past_part = [int(i) for i in csv_results["part"].tolist() if not math.isnan(i)]

    else:
        past_iteration = 0
        past_part = []

    if (not args.part) or (args.part > past_iteration) or (int(args.part) not in past_part) or ((args.imbalance_generator) and (int(args.seed) not in past_seed)):
        if not args.wandboff:
            wandb.init(project=args.project_name)

        device = torch.device("cuda:" + (args.gpu_no if torch.cuda.device_count() > 1 else "0") if (torch.cuda.is_available() and not args.debug) else "cpu")

        args.device = device
        set_seed(args)

        if (args.num_class == 3):
            args.ID_TO_LABEL = en3ID_TO_LABEL
            args.LABEL_TO_ID = DICT_LABELS["bio"]["LABEL_TO_en3ID"]
        else:
            args.ID_TO_LABEL = {value: key for key, value in DICT_LABELS["conll" if args.corpus in ["conll"] else "ontonotes5"]["LABEL_TO_ID"].items()}
            args.LABEL_TO_ID = DICT_LABELS["conll" if args.corpus in ["conll"] else "ontonotes5"]["LABEL_TO_ID"]

        tokenizer = AutoTokenizer.from_pretrained(args.model_name_or_path)
        train_file = os.path.join(args.data_dir, "train.txt")

        if args.imbalance_generator:
            train_features, num_iter = read_conll(
                args, train_file, tokenizer, args.num_class, max_seq_length=args.max_seq_length,
                num_training_sentences=args.num_training_sentences, imbalance_generator=args.imbalance_generator,
                generate_imbRatio=args.generate_imbRatio, generate_imbRatio_error=args.generate_imbRatio_error, randMaxSeeds=args.randMaxSeeds
            )
        else:
            train_features, num_iter = read_conll(
                args, train_file, tokenizer, args.num_class, max_seq_length=args.max_seq_length,
                imbalance_generator=args.imbalance_generator,
                generate_imbRatio=args.generate_imbRatio, generate_imbRatio_error=args.generate_imbRatio_error, randMaxSeeds=args.randMaxSeeds
            )

        args.data_format = "conll" if args.corpus in ["conll"] else "bio"

        num_tokens = 0
        num_en = 0
        
        for i in range(args.num_class):
            args.__setattr__("num_" + args.ID_TO_LABEL[i], 0)
            args.__setattr__("pos_len_" + args.ID_TO_LABEL[i], 0)

        for proc_instance in train_features:
            num_tokens += sum([1 if i!=-1 else 0 for i in proc_instance["labels_en2"]])
            num_en += sum([1 if i==1 else 0 for i in proc_instance["labels_en2"]])

            for i in range(args.num_class):
                args.__setattr__("num_" + args.ID_TO_LABEL[i], args.__getattribute__(
                    "num_" + args.ID_TO_LABEL[i]) + sum([1 if j==i else 0 for j in proc_instance["labels"]]
                ))
                args.__setattr__("pos_len_" + args.ID_TO_LABEL[i], args.__getattribute__(
                    "pos_len_" + args.ID_TO_LABEL[i]) + sum([1 if j==i else 0 for j in proc_instance["labels"]]
                ))
            
        print("num_tokens:", num_tokens)

        print("num_en:", num_en)
        ratio_en = float(num_en/num_tokens)
        print("num_en/num_tokens:", ratio_en)

        args.imratio_en = ratio_en

        for i in range(args.num_class):
            class_name = args.ID_TO_LABEL[i]
            print("num_" + class_name  + ":", args.__getattribute__("num_" + class_name))
            ratio_i = float(args.__getattribute__("num_" + class_name)/num_tokens)
            print("ratio_" + class_name + ":", ratio_i)
            args.__setattr__("imratio_" + class_name, ratio_i)

        dev_file = os.path.join(args.data_dir, "dev.txt")
        dev_features,_ = read_conll(args, dev_file, tokenizer, args.num_class, max_seq_length=args.max_seq_length)
        test_file = os.path.join(args.data_dir, "test.txt")
        test_features,_ = read_conll(args, test_file, tokenizer, args.num_class, max_seq_length=args.max_seq_length)

        args.run_name = "%s%s_size%s_part%s%s%s" % (
            "bart_" if args.model_name_or_path == "facebook/bart-base" else ("biobert_" if args.model_name_or_path == "dmis-lab/biobert-base-cased-v1.1" else ""),
            prefix,
            args.num_training_sentences if args.num_training_sentences else "",
            args.part if args.part else "",
            "_seed_%s"%str(args.seed) if args.imbalance_generator else "",
            '_imb_ratio_%s'%str(args.generate_imbRatio) if args.imbalance_generator else ''
        )

        if not args.wandboff:
            wandb.run.name = args.run_name

        # Change this to remember the logits, don't save the models to args.output_dir
        args.output_dir = os.path.join(args.output_dir, args.run_name)
        if not os.path.exists(args.output_dir):
            os.makedirs(args.output_dir)

        benchmarks = (
            ("dev", dev_features),
            ("test", test_features),
        )

        if (args.aucopt) or (args.comopt) or (args.ceopt):
            if args.ensembleopt:
                ensembleLogits = []
                
                for i in range(args.num_ensembles):
                    model = NEROVAModel(args)
                    cur_model = copy.deepcopy(model)
                    del model

                    if args.meta_learning:
                        best_test_output, best_dev_output, Loss, bestlogits = trainM(args, cur_model, train_features, benchmarks, tokenizer)
                    else:
                        best_test_output, best_dev_output, Loss, bestlogits = trainOVA(args, cur_model, train_features, benchmarks, tokenizer)
                    
                    saveEnsembles(bestlogits, args, i, benchmarks)
                    ensembleLogits.append(bestlogits)
                    # ## 6542 is a random seed to compare with all methods
                    args.seed = args.seed + 6542
                    set_seed(args)

                best_test_output, best_dev_output = reportsFromEnsembles(ensembleLogits, args, benchmarks)
            
            else:
                model = NEROVAModel(args)
                
                if args.meta_learning:
                    best_test_output, best_dev_output, Loss, _ = trainM(args, model, train_features, benchmarks, tokenizer)
                else:
                    best_test_output, best_dev_output, Loss, _ = trainOVA(args, model, train_features, benchmarks, tokenizer)

        else:
            if args.ensembleopt:
                ensembleLogits = []
                
                for i in range(args.num_ensembles):
                    model = NERModel(args)
                    cur_model = copy.deepcopy(model)
                    del model

                    _, _, bestlogits = train(args, cur_model, train_features, benchmarks, tokenizer)

                    saveEnsembles(bestlogits, args, i, benchmarks)
                    ensembleLogits.append(bestlogits)
                    ## 6542 is a random seed to compare with all methods
                    args.seed = args.seed + 6542
                    set_seed(args)

                best_test_output, best_dev_output = reportsFromEnsembles(ensembleLogits, args, benchmarks)
            else:
                model = NERModel(args)
                best_test_output, best_dev_output, _ = train(args, model, train_features, benchmarks, tokenizer)

        results = {
            "bert": args.model_name_or_path,
            "optimizertype": args.optimizertype,
            "num_training_sentences": args.num_training_sentences,
            "part": args.part,
            "learning_rate_auc": args.learning_rate_auc,
            "learning_rate": args.learning_rate,
            "imratio_en": args.imratio_en
        }

        for class_idx in range(args.num_class):
            class_name = args.ID_TO_LABEL[class_idx]
            results["imratio_" + class_name] = args.__getattribute__("imratio_" + class_name)

        for report in [best_dev_output, best_test_output]:
            for key in report:
                results[key] = report[key]

        if args.imbalance_generator:
            results["generate_imbRatio"] = args.generate_imbRatio
            results["generate_imbRatio_error"] = args.generate_imbRatio_error
            results["randMaxSeeds"] = args.randMaxSeeds
            results["seed"] = args.seed
            results["imratio_en"] = args.imratio_en
            results["num_iter"] = num_iter

        while True:
            try:
                # Getting the path to the csv file and get the dataframe
                if os.path.exists(results_path):
                    csv_file = pd.read_csv(results_path, index_col=False)
                else:
                    csv_file = pd.DataFrame()
                
                # Appending the dictionary to the csv file and write it out!
                csv_file = csv_file.append(results, ignore_index=True)
                csv_file.to_csv(results_path, index=False)
                
                break
            
            except:
                time.sleep(random.randint(0, 9))
                exit(0)

        gc.collect()  # release all from the memory


if __name__ == "__main__":
    main()
