import torch
import numpy as np
import random, os, logging, time, pdb
import seqeval.metrics
from tqdm import tqdm
from torch.utils.data import DataLoader
from itertools import chain
import itertools
from itertools import product


def set_seed(args):
    random.seed(args.seed)
    np.random.seed(args.seed)
    torch.manual_seed(args.seed)
    torch.cuda.manual_seed_all(args.seed)


def collate_fn(batch):
    max_len = max([len(f["input_ids"]) for f in batch])
    input_ids = [f["input_ids"] + [0] * (max_len - len(f["input_ids"])) for f in batch]
    attention_mask = [[1.0] * len(f["input_ids"]) + [0.0] * (max_len - len(f["input_ids"])) for f in batch]
    labels = [f["labels"] + [-1] * (max_len - len(f["labels"])) for f in batch]
    input_ids = torch.tensor(input_ids, dtype=torch.long)
    attention_mask = torch.tensor(attention_mask, dtype=torch.float)
    labels = torch.tensor(labels, dtype=torch.long)
    labels_en3 = [f["labels_en3"] + [-1] * (max_len - len(f["labels_en3"])) for f in batch]
    labels_en3 = torch.tensor(labels_en3, dtype=torch.long)
    labels_en2 = [f["labels_en2"] + [-1] * (max_len - len(f["labels_en2"])) for f in batch]
    labels_en2 = torch.tensor(labels_en2, dtype=torch.long)
    labels_be2 = [f["labels_be2"] + [-1] * (max_len - len(f["labels_be2"])) for f in batch]
    labels_be2 = torch.tensor(labels_be2, dtype=torch.long)

    pos_dict = {}

    output = {
        'input_ids': input_ids,
        'attention_mask': attention_mask,
        'labels': labels,
        "labels_en3": labels_en3,
        "labels_en2": labels_en2,
        "labels_be2": labels_be2,
    }
    return {**output, **pos_dict}


def saveBestLogits(args, bestlogits, label_set=None):
    if label_set:
        combinebestlogits = {}
        bestlogitsdev = [None]*args.num_class
        bestlogitstest = [None]*args.num_class
        for label, label_logits_dev_test in bestlogits.items():
            for benchmark, benchmark_logits in label_logits_dev_test.items():
                path = os.path.join(args.output_dir, label, benchmark)
                if not os.path.exists(path):
                    os.makedirs(path)
                for class_idx, class_logits in enumerate(benchmark_logits):
                    if class_logits:
                        file = os.path.join(path, str(class_idx) + '.pt')
                        torch.save(torch.tensor(class_logits, dtype=torch.float32), file)
                        if benchmark == 'dev_logits':
                            bestlogitsdev[class_idx] = torch.tensor(class_logits)
                        else:
                            bestlogitstest[class_idx] = torch.tensor(class_logits)
        
        combinebestlogits["dev"] = bestlogitsdev
        combinebestlogits["test"] = bestlogitstest

        return combinebestlogits
                
    else:
        output_dir = args.output_dir
        for name, logitsLi in bestlogits.items():
            path = os.path.join(output_dir, name)
            if not os.path.exists(path):
                os.makedirs(path)
            if(args.crfopt):
                file = os.path.join(path, name + '.pt')
                torch.save(torch.tensor(logitsLi), file)
            else:
                for batch_idx, logits in enumerate(logitsLi):
                    file = os.path.join(path, name + str(batch_idx) + '.pt')
                    torch.save(logits.to(dtype=torch.float32), file)


def getBestLogitsFromDict(args, bestlogits, label_set=None):
    bestlogitsdev = [None]*args.num_class
    bestlogitstest = [None]*args.num_class
    
    for label in label_set:
        label_logits_dev_test = bestlogits[label]
        for benchmark, benchmark_logits in label_logits_dev_test.items():
            for class_idx, class_logits in enumerate(benchmark_logits):
                if class_logits:
                    if benchmark == 'dev_logits':
                        bestlogitsdev[class_idx] = class_logits
                    else:
                        bestlogitstest[class_idx] = class_logits
    
    return bestlogitsdev, bestlogitstest


def getBestLogits(args, label_set):
    bestlogitsdev = [None]*args.num_class
    bestlogitstest = [None]*args.num_class
    
    for label in label_set:
        for benchmark in ['dev_logits', 'test_logits']:
            path = os.path.join(args.output_dir, label, benchmark)
            assert os.path.exists(path)
            for class_idx in range(args.num_class):
                if os.path.exists(os.path.join(path, str(class_idx) + '.pt')):
                    file = os.path.join(path, str(class_idx) + '.pt')
                    if benchmark == 'dev_logits':
                        bestlogitsdev[class_idx] = torch.load(file)
                    else:
                        bestlogitstest[class_idx] = torch.load(file)
    
    return bestlogitsdev, bestlogitstest


def get_logger(args):
    savedir = os.path.join(args.output_dir, 'logs')
    log_file = args.log_file

    if not os.path.exists(savedir):
        os.makedirs(savedir)

    if len(log_file) == 0:
        log_file = os.path.join(savedir, time.asctime())
        args.log_file = log_file

    logger = logging.getLogger()
    logging.basicConfig(
        format="%(asctime)s %(levelname)s: %(message)s",
        datefmt="%d-%m-%Y %H:%M:%S",
        level=logging.INFO,
        filename=log_file
    )
    return logger


def saveEnsembles(bestlogits, args, ensemble_idx, benchmarks):
    for tag, _ in benchmarks:
        path = os.path.join(args.output_dir, "ensemble", str(ensemble_idx), tag)
        if not os.path.exists(path):
            os.makedirs(path)
        for classOrBatch_idx, labellogit in enumerate(bestlogits[tag]):
            file = os.path.join(path, str(classOrBatch_idx) + '.pt')
            torch.save(torch.tensor(labellogit, dtype=torch.float32), file)


def evaluateByBestLogits(ensembleLogits, args, dataloader, tag):
    preds, keys = [], []
    ensLogitsLi = []

    for batch in dataloader:
        batch = {key: value.to(args.device) for key, value in batch.items()}
        keys += batch['labels'].cpu().numpy().flatten().tolist()
        # mask = ~torch.eq(batch["labels"], -1)

    for ensemble_idx, bestlogits in enumerate(ensembleLogits):
        if args.crfopt:
            ## items are predictions
            ensLogitsLi.append(torch.tensor(bestlogits[tag], dtype=torch.float32))
        elif args.onevsall:
            # pdb.set_trace()
            logitsLi = bestlogits[tag]
            ensLogitsLi.append(torch.stack(logitsLi, dim=0))
        else:
            ## items are a list of seq * class logits
            logitsLi = bestlogits[tag]
            ensLogitsLi.append(torch.cat(logitsLi, dim=0))

    if args.crfopt:
        keys = [key for key in keys if key != -1]
        stackLogits = torch.stack(ensLogitsLi, dim=0)
        aveLogits = torch.mean(stackLogits, dim=0)
        preds = torch.where(aveLogits > 0.5, 1, 0).cpu().detach().numpy().flatten().tolist()
    elif args.onevsall:
        aveLogits = sum(ensLogitsLi)/len(ensLogitsLi)
        preds = np.argmax(aveLogits.cpu().numpy(), axis=0).tolist()
    else:
        aveLogits = torch.stack(ensLogitsLi).mean(dim=0)
        preds = np.argmax(aveLogits.cpu().numpy(), axis=-1).tolist()

    preds, keys = list(zip(*[[pred, key] for pred, key in zip(preds, keys) if key != -1]))
    preds = [args.ID_TO_LABEL[pred] for pred in preds]
    keys = [args.ID_TO_LABEL[key] for key in keys]

    f1 = seqeval.metrics.f1_score([keys], [preds], scheme="IOB2")  # IOB2 works for BIO
    precision = seqeval.metrics.precision_score([keys], [preds], scheme="IOB2")
    recall = seqeval.metrics.recall_score([keys], [preds], scheme="IOB2")

    result = seqeval.metrics.classification_report([keys], [preds], digits=4, scheme="IOB2")
    output = {
        tag + "_f1": f1,
        tag + "_precision": precision,
        tag + "_recall": recall,
    }

    return output, result


def reportsFromEnsembles(ensembleLogits, args, benchmarks):

    batch_size = 128
    try:
        results = findtheoptimalensembles(ensembleLogits, args, benchmarks, batch_size)
    except:
        results = {}
        for tag, features in benchmarks:
            dataloader = DataLoader(features, batch_size=128, shuffle=False, collate_fn=collate_fn, drop_last=False)
            result, _ = evaluateByBestLogits(ensembleLogits, args, dataloader, tag)
            results[tag] = result

    return results['test'], results['dev']


def generate_valid_logits3(args, benchmarks, batch_size, collate_fn=collate_fn):
  
    newlogit3dic={}
    for tag, feature in benchmarks:
        maskLi, true_labels = [], []
        eval_dataloader = DataLoader(feature, batch_size=batch_size, shuffle=False,
                                                    collate_fn=collate_fn, drop_last=False)
        for _, batch in enumerate(eval_dataloader):
            batch = {key: value.to(args.device) for key, value in batch.items()}
            with torch.no_grad():
                labels = batch['labels'].reshape(-1,)
                mask = ~torch.eq(labels, -1)
                labels = labels[~labels.eq(-1)]

            maskLi += mask.cpu().detach().numpy().flatten().tolist()
            true_labels += labels.cpu().detach().numpy().flatten().tolist()
        maskzeros = torch.tensor(maskLi).unsqueeze(1)

        newlogit3dic[tag] = (maskzeros, true_labels)
    return newlogit3dic

#(ensembleLogits, args, dataloader, tag)
def findtheoptimalensembles(logitsli, args, benchmarks, batch_size):
    
    reshapeLogits = []
    logits3dic = {}
    best_score = -1
    flag = True
    results = {}

    ensbestlogitsdev = []
    ensbestlogitstest = []

    for eachModel in logitsli:
        if (args.onevsall):
            for k,v in eachModel.items():
                if k == 'dev':
                    ensbestlogitsdev.append(torch.stack(v, dim=0).transpose(0, 1))
                if k == 'test':
                    ensbestlogitstest.append(torch.stack(v, dim=0).transpose(0, 1))
        else:
            for k,v in eachModel.items():
                if k == 'dev':
                    ensbestlogitsdev.append(torch.cat(v, dim=0))
                if k == 'test':
                    ensbestlogitstest.append(torch.cat(v, dim=0))

    print(len(ensbestlogitsdev))

    for devlogit, testlogit in zip (ensbestlogitsdev, ensbestlogitstest):
        reshapeLogits.append((devlogit, testlogit))

    for combEle in range(len(reshapeLogits) + 1):
        for subLogits in itertools.combinations(reshapeLogits, combEle):
            if len(subLogits) != 0:

                logits3dic['dev'] = sum([x[0] for x in subLogits]) / len(subLogits)
                logits3dic['test']  = sum([x[1] for x in subLogits]) / len(subLogits)

                if (flag):
                    newMaskLabelDic = generate_valid_logits3(args, benchmarks, batch_size)
                    flag = False

                cur_f1, bestDevRes = calculate_f1(newMaskLabelDic, 'dev', logits3dic, args)

                if (cur_f1 > best_score) :
                    best_score = cur_f1
                    _, bestTestRes = calculate_f1(newMaskLabelDic, 'test', logits3dic, args)
                    results['dev'] = bestDevRes
                    results['test'] = bestTestRes

    return results


def calculate_f1(newMaskLabelDic, tag, logits3dic, args):
    mask = newMaskLabelDic[tag][0]
    true_labels = newMaskLabelDic[tag][1]
    masked = logits3dic[tag]  * mask
    masked = masked[~masked.eq(0)]
    logitsDev = masked.reshape(-1, args.num_class)
    logit_preds = torch.argmax(logitsDev,-1).cpu().detach().numpy().flatten().tolist()
    preds, keys = list(zip(*[[pred, key] for pred, key in zip(logit_preds, true_labels) if key != -1]))
    preds = [args.ID_TO_LABEL[pred] for pred in preds]
    keys = [args.ID_TO_LABEL[key] for key in keys]
    cur_f1 = seqeval.metrics.f1_score([keys], [preds], scheme="IOB2")
    precision = seqeval.metrics.precision_score([keys], [preds], scheme="IOB2")
    recall = seqeval.metrics.recall_score([keys], [preds], scheme="IOB2")
    output = {
        tag + "_f1": cur_f1,
        tag + "_precision" : precision,
        tag + "_recall": recall,
    }
    return cur_f1, output